! *******************************************************************
! *       SENDFILE send a file to the client                        *
! *-----------------------------------------------------------------*
! * Required subroutines:                  FILEERR.INC              *
! *                                        NETLNK.SR                *
! *                                                                 *
! *-----------------------------------------------------------------*
! * Author: W. Gerth 24.03.2000                                     *
! *-----------------------------------------------------------------*
! * Last modification: 21.02.2005 J.Bartels                         *
! *   File operations synchronized for simultaneous access          *
! *   new functions adapted from STARTUP.P                          *
! *                                                                 *
! *******************************************************************

/****************/
/* sema_request */
/****************/
file_sema_request: PROC ( nr FIXED ) ;
   REQUEST GLB_FILE_WAIT( nr );
END;
/****************/
/* sema_release */
/****************/
file_sema_release: PROC ( nr FIXED ) ;
   RELEASE GLB_FILE_WAIT( nr );
   GLB_FILE_NAMES(nr).cnt = GLB_FILE_NAMES(nr).cnt - 1;
   IF GLB_FILE_NAMES(nr).cnt EQ 0 THEN
      GLB_FILE_NAMES(nr).t = ' ';
      RELEASE GLB_FILE_FREE;
   FIN;
END;
/****************/
/* file_sema_set_all */
/****************/
file_sema_set_all: PROC ;
   FOR i TO (1 UPB GLB_FILE_WAIT) REPEAT
     SEMASET 1,GLB_FILE_WAIT(i);
   END;

   SEMASET 1,GLB_FILE_SEMA        ;
   SEMASET MAX_USER,GLB_FILE_FREE ;

   FOR i TO (1 UPB GLB_FILE_NAMES) REPEAT
      GLB_FILE_NAMES(i).t = ' ';
      GLB_FILE_NAMES(i).cnt = 0;
   END;

END;

/**************/
/*  ins_file  */
/**************/

ins_file: PROC(name CHAR(128));

   /*---------------------------------------------------------------*/
   /* File in die Liste der offenen Files eintragen und ggf. auf    */
   /* einen freien Platz warten. Ist ein Eintrag schon vorhanden    */
   /* wird auf das Schliessen des entsprechenden Files gewartet.    */
   /*---------------------------------------------------------------*/

   DCL free FIXED INIT(0);
   DCL found FIXED INIT(0);

   REQUEST GLB_FILE_FREE ;
   REQUEST GLB_FILE_SEMA;    /* darf nur einer an der Tabelle fummeln    */;
   FOR i TO ( 1 UPB GLB_FILE_NAMES ) WHILE found EQ 0 REPEAT
      IF GLB_FILE_NAMES(i).cnt EQ 0 THEN
         free = i;
      ELSE
         IF GLB_FILE_NAMES(i).t EQ name THEN
            found = i;
         FIN;
      FIN;
   END;
   IF found NE 0 THEN               /* bitte warten  ...            */;
      GLB_FILE_NAMES( found ).cnt = GLB_FILE_NAMES( found ).cnt  + 1 ;
      RELEASE GLB_FILE_SEMA;             /* andere sollen auch mal       */;
      RELEASE GLB_FILE_FREE ;
      #IF MAXILOG+MINILOG;
        PUT 'ins_file: Waiting for ',name TO HTPD_LOGLINK;
      #FIN;
      CALL file_sema_request(found);     /* hier wird gewartet           */;
      #IF MAXILOG+MINILOG;
        PUT 'ins_file: Resuming on ',name TO HTPD_LOGLINK;
      #FIN;
   ELSE
      IF free NE 0 THEN             /* Eintragen                    */;
         GLB_FILE_NAMES(free).t = name;
         GLB_FILE_NAMES(free).cnt = GLB_FILE_NAMES(free).cnt + 1;
         CALL file_sema_request(free);
         RELEASE GLB_FILE_SEMA;
      ELSE                          /* sollte nicht passieren       */;
         RELEASE GLB_FILE_SEMA;
         WRITE 'ins_file: PANIC'\0D0A\'' TO HTPD_Alarm_out;
      FIN;
   FIN;
END; /* PROC ins_file ----------------------------------------------*/;

/**************/
/*  del_file  */
/**************/

del_file: PROC(name CHAR(128));

   /*---------------------------------------------------------------*/
   /* Es wird eine Datei aus der Liste genommen und ggf. die        */
   /* Semaphore freigegeben.                                        */
   /*---------------------------------------------------------------*/

   DCL found FIXED INIT(0);

   REQUEST GLB_FILE_SEMA;
   FOR i TO ( 1 UPB GLB_FILE_NAMES )WHILE found EQ 0 REPEAT
      IF GLB_FILE_NAMES(i).t EQ name THEN
         found = i;
      FIN;
   END;
   IF found > 0
   THEN
      CALL file_sema_release(found);
   FIN;
   RELEASE GLB_FILE_SEMA;

END; /* PROC del_file ----------------------------------------------*/

SENDFILE: PROC( CHILD FIXED,
                FILEINFO FILEPACK IDENT,
                stderr DATION OUT ALPHIC IDENT,
                loglink DATION OUT ALPHIC IDENT
              );

   DCL BUFFER(1000) CHAR(1);     ! The internal buffer
   DCL SCRATCH REF DATION INOUT ALPHIC;
   DCL FILESTATE             FIXED(15);
   DCL ILEN                  FIXED(31);  ! Incoming data len
   DCL COUNTER               FIXED(31);  ! total no of bytes

   DCL NETERROR              FIXED(15)     INIT(0);
   DCL NETST                 FIXED;
   DCL Fname REF CHAR(255);              ! For fast access only
   DCL Fnamelen FIXED;

   IF FILEINFO.Status /= F_OPEN
   THEN
     #IF MAXILOG
       WRITE 'S: SENDFILE file not open.' TO loglink;
     #FIN;
     RETURN;
   FIN;

   #IF MAXILOG
     WRITE 'S: SENDFILE starting.' TO loglink;
   #FIN;
   SCRATCH = FILEINFO.DatPTR;    ! For faster access only

   ILEN = 1 UPB BUFFER; COUNTER = 0(31);

   WHILE ILEN > 0(31) REPEAT
         ILEN = 1 UPB BUFFER;
         READ BUFFER FROM SCRATCH BY S(ILEN);
         FILESTATE:=ST(SCRATCH);
         IF ILEN > 0(31) THEN
           NETST=NETLINK(CHILD,BUFFER(1).CHAR(1),ILEN FIT 1(15),SO_WRITE,FLG_PUSH+FLG_WAIT,stderr,loglink);
           NETERROR=NETSTERR(NETST,'1'B,stderr);
           IF NETERROR NE 0 AND NETERROR NE 3 THEN
             #IF SICOLOG;
             #ELSE;
             WRITE 'HTTP-USER: NETERROR in SENDFILE SO_WRITE'\0D0A\''
                   TO stderr;
             #FIN;
             #IF MAXILOG+MINILOG
               WRITE 'S: NETERROR in SENDFILE SO_WRITE.' TO loglink;
               NETERROR=NETSTERR(NETST,'1'B,loglink);
             #FIN;
             ILEN = 0(31);
           FIN; /* IF NETERROR NE 0 THEN */
           COUNTER = COUNTER+ILEN;
         ELSE
           IF FILESTATE GT 1 THEN
              #IF MAXILOG+MINILOG
               WRITE 'S: READ-Error in SENDFILE.' TO loglink;
               FileErr(FILESTATE,loglink);
              #FIN;
           FIN; ! if FILESTATE ...
         FIN;  ! if ILEN > 0;
   END; ! WHILE REPEAT

! .................... Leaving Sendfile .......................
   CLOSE SCRATCH;
   FILESTATE:=ST(SCRATCH);
   IF (FILESTATE NE 0) THEN
      Fname = FILEINFO.Name; Fnamelen=LEN(Fname);
      #IF SICOLOG;
        PUT GetLogfileCode( LC_CLOSEERR ), Fname, FILESTATE TO stderr BY LIST,X,A(Fnamelen),X,LIST,SKIP;
      #ELSE;
        PUT 'HTTP-USER: SENDFILE unable to close file ',Fname
            TO stderr BY A,A(Fnamelen),SKIP;
        PUT 'HTTP-USER: File-error ',FILESTATE TO stderr
             BY SKIP,A,F(3);
        FileErr(FILESTATE,stderr);
      #FIN;
      #IF MAXILOG+MINILOG
        PUT 'S: SENDFILE unable to close file ',Fname
              TO loglink BY A,A(Fnamelen);
        PUT 'S: File-error: ',FILESTATE TO loglink  BY A,F(3);
        FileErr(FILESTATE,loglink);
      #FIN;
   FIN;

   #IF MAXILOG
      PUT 'S: ',COUNTER,' bytes sent.' TO
           loglink BY A,F(10),A;
      WRITE 'S: SENDFILE terminated.' TO loglink;
   #FIN;! PRAEPROZESSOR-BEFEHL !

   CALL del_file( FILEINFO.Name );

END; /*SENDFILE*/

